@@ -84,11 +84,11 @@ def fit_generator_autosized(model,
     # prepare callbacks
     model.history = cbks.History()
     _callbacks = [cbks.BaseLogger(
-        stateful_metrics=model.stateful_metric_names)]
+        stateful_metrics=model.metrics_names[1:])]
     # instead of ProgbarLogger (but only for first epoch):
     if verbose:
         print('Epoch 1/%d' % epochs)
-        progbar = Progbar(target=None, verbose=1, stateful_metrics=model.stateful_metric_names)
+        progbar = Progbar(target=None, verbose=1, stateful_metrics=model.metrics_names[1:])
     _callbacks += (callbacks or []) + [model.history]
     callbacks = cbks.CallbackList(_callbacks)
 
@@ -155,8 +155,7 @@ def fit_generator_autosized(model,
         # Construct epoch logs.
         epoch_logs = {}
         while epoch < epochs:
-            for m in model.stateful_metric_functions:
-                m.reset_states()
+            model.reset_metrics()
             callbacks.on_epoch_begin(epoch)
             steps_done = 0
             batch_index = 0
@@ -197,7 +196,8 @@ def fit_generator_autosized(model,
                 
                 outs = model.train_on_batch(x, y,
                                             sample_weight=sample_weight,
-                                            class_weight=class_weight)
+                                            class_weight=class_weight,
+                                            reset_metrics=False)
                 
                 if not isinstance(outs, list):
                     outs = [outs]
@@ -217,7 +217,10 @@ def fit_generator_autosized(model,
                 
                 if callback_model.stop_training:
                     break
-            
+
+            if steps_done == 0:
+                raise ValueError('Output of generator must not be empty')
+
             if epoch == initial_epoch:
                 if verbose:
                     log_values = []
@@ -277,7 +280,10 @@ def fit_generator_autosized(model,
             epoch += 1
             if callback_model.stop_training:
                 break
-    
+
+            if is_sequence and workers == 0:
+                generator.on_epoch_end()
+
     finally:
         try:
             if enqueuer is not None:
@@ -299,16 +305,7 @@ def evaluate_generator_autosized(model, generator,
     """See docstring for `Model.evaluate_generator`."""
     model._make_test_function()
     
-    stateful_metric_indices = []
-    if hasattr(model, 'metrics'):
-        for m in model.stateful_metric_functions:
-            m.reset_states()
-        stateful_metric_indices = [
-            i for i, name in enumerate(model.metrics_names)
-            if str(name) in model.stateful_metric_names]
-    else:
-        stateful_metric_indices = []
-    
+    model.reset_metrics()
     callbacks = cbks.CallbackList(callbacks or [])
     
     # it's possible to callback a different model than self:
@@ -407,7 +404,9 @@ def evaluate_generator_autosized(model, generator,
             batch_logs['size'] = batch_size
             callbacks.on_batch_begin(steps_done, batch_logs)
             
-            outs = model.test_on_batch(x, y, sample_weight=sample_weight)
+            outs = model.test_on_batch(x, y,
+                                       sample_weight=sample_weight,
+                                       reset_metrics=False)
             if not isinstance(outs, list):
                 outs = [outs]
             for l, o in zip(model.metrics_names, outs):
@@ -431,13 +430,9 @@ def evaluate_generator_autosized(model, generator,
         if enqueuer is not None:
             enqueuer.stop()
     
-    averages = []
-    for i in range(len(model.metrics_names)):
-        if i not in stateful_metric_indices:
-            averages.append(np.average([out[i] for out in outs_per_batch],
-                                       weights=batch_sizes))
-        else:
-            averages.append(float(outs_per_batch[-1][i]))
+    averages = [float(outs_per_batch[-1][0])]  # index 0 = 'loss'
+    for i in range(1, len(outs)):
+        averages.append(np.float64(outs_per_batch[-1][i]))
     if len(averages) == 1:
         return averages[0], steps_done
     return averages, steps_done
